"""
Tensorboard support.
"""

import click

from ..config import Config
from ..cli_utils import OrderedGroup, option_hostname
from .cli_utils import option_port, option_open_browser
from .remote_server import RemoteServer


class TensorBoard(RemoteServer):
    """
    TensorBoard remote server.
    """

    def __init__(
        self,
        port: int,
        hostname: str,
        logdir: str = "",
        open_browser: bool = True,
    ) -> None:
        super().__init__(
            server_start_command=f"tensorboard --bind_all --port {port} --logdir {logdir}",
            server_stop_command="pkill tensorboard",
            port=port,
            hostname=hostname,
            open_browser=open_browser,
            local_url=f"http://localhost:{port}",
        )

    def _server_start_callback(self, server_output_line: str) -> bool:
        return server_output_line.startswith("TensorBoard")


#######
# CLI #
#######
_option_tensorboard_port = option_port(service_name="tensorboard")


# pylint: disable=unused-argument
def _callback_option_tensorboard_logdir(ctx, param, value) -> str:
    if value:
        return value

    return Config().remote_servers.tensorboard.logdir


_option_tensorboard_logdir = click.option(
    "-d",
    "--logdir",
    callback=_callback_option_tensorboard_logdir,
    type=str,
    help="Log directory.",
)


@click.group(cls=OrderedGroup, invoke_without_command=True)
@click.pass_context
@_option_tensorboard_port
@option_hostname
@_option_tensorboard_logdir
@option_open_browser
def tensorboard(
    ctx,
    port: int,
    hostname: str,
    logdir: str,
    open_browser: bool,
) -> None:
    """
    Run TensorBoard on the remote host.
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            tensorboard_run,
            port=port,
            hostname=hostname,
            logdir=logdir,
            open_browser=open_browser,
        )


@tensorboard.command("run")
@_option_tensorboard_port
@option_hostname
@option_open_browser
@_option_tensorboard_logdir
def tensorboard_run(
    port: int,
    hostname: str,
    logdir: str,
    open_browser: bool,
) -> None:
    """
    Run TensorBoard on the remote host.
    """
    TensorBoard(
        hostname=hostname,
        port=port,
        logdir=logdir,
        open_browser=open_browser,
    ).start()


@tensorboard.command("stop")
@_option_tensorboard_port
@option_hostname
def tensorboard_stop(port: int, hostname: str) -> None:
    """
    Stop TensorBoard on the remote host.
    """
    TensorBoard(
        hostname=hostname,
        port=port,
    ).stop_server()



"""
Utility functions for interacting with oarsub.
"""

from datetime import datetime
from os.path import join

from ...config import Config, Oarsub
from ...user_interaction import to_bold, print_info


def print_cluster_request_prop(
    property_name: str,
    value: str,
) -> None:
    num_dots: int = 25 - len(property_name)
    to_print: str = "\t" + to_bold(property_name + ":") + "." * num_dots + str(value)
    print_info(to_print)


def _print_cluster_request(
    oarsub_config: Oarsub,
    host_id: str,
    job_name: str,
) -> None:
    print_info(to_bold("Cluster request:"))
    print_cluster_request_prop(
        property_name="Job name",
        value=job_name,
    )
    print_cluster_request_prop(
        property_name="Number of hosts",
        value=oarsub_config.num_hosts,
    )
    if oarsub_config.num_cpu_cores > 0:
        print_cluster_request_prop(
            property_name="Number of CPU cores",
            value=oarsub_config.num_cpu_cores,
        )
    print_cluster_request_prop(
        property_name="Number of GPUs",
        value=oarsub_config.num_gpus,
    )
    print_cluster_request_prop(
        property_name="Walltime",
        value=oarsub_config.walltime,
    )
    print_cluster_request_prop(
        property_name="Cluster name",
        value=oarsub_config.cluster_name,
    )
    if host_id != "":
        print_cluster_request_prop(
            property_name="Host ID",
            value=host_id,
        )
    if oarsub_config.custom_property_query:
        print_cluster_request_prop(
            property_name="Custom property query",
            value=oarsub_config.custom_property_query,
        )
    print_cluster_request_prop(
        property_name="Besteffort",
        value=oarsub_config.besteffort,
    )
    print_cluster_request_prop(
        property_name="Idempotent",
        value=oarsub_config.idempotent,
    )


def get_oarsub_cmd_prefix(
    host_id: str,
    job_name: str,
) -> list[str]:
    # Get the oarsub config
    oarsub_config: Oarsub = Config().oarsub

    # Oarsub command
    oarsub_cmd: list[str] = ["oarsub"]

    # Name the job
    oarsub_cmd += ["--name", job_name]

    # Resource list
    resource_list: str = f"/host={oarsub_config.num_hosts}"
    if oarsub_config.num_cpu_cores > 0:
        resource_list += f"/core={oarsub_config.num_cpu_cores}"

    if oarsub_config.num_gpus > 0:
        resource_list += f"/gpudevice={oarsub_config.num_gpus}"

    if oarsub_config.walltime != "":
        resource_list += f",walltime={oarsub_config.walltime}"

    oarsub_cmd += ["-l", resource_list]

    # Check that the cluster name is valid
    assert oarsub_config.cluster_name in [
        "beagle",
        "cp",
        "kinovis",
        "mistis",
        "nanod",
        "perception",
        "thoth",
    ], f"Unknown cluster name: {oarsub_config.cluster_name}"

    # Property list
    property_list: str
    # Using directly the provided property query, ignoring other options.
    if oarsub_config.custom_property_query:
        property_list = '"' + oarsub_config.custom_property_query + '"'

    # Using the other config settings.
    else:
        property_list = f"\"cluster='{oarsub_config.cluster_name}'"

        if host_id:
            assert 4 <= len(host_id) <= 5
            assert host_id.startswith("gpu") or host_id.startswith("node")
            cluster_host_name: str = f"{host_id}-{oarsub_config.cluster_name}.inrialpes.fr"

            property_list += f" AND host='{cluster_host_name}'"

        property_list += '"'
    oarsub_cmd += ["-p", property_list]

    if oarsub_config.besteffort:
        oarsub_cmd += ["-t", "besteffort"]

    if oarsub_config.idempotent:
        oarsub_cmd += ["-t", "idempotent"]

    _print_cluster_request(
        oarsub_config=oarsub_config,
        job_name=job_name,
        host_id=host_id,
    )

    return oarsub_cmd


def add_standard_output_params(
    oarsub_log_path: str,
    base_name: str = "",
) -> list[str]:
    time_stamp: str = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    if not base_name:
        base_name = f"OAR.{time_stamp}.%jobid%.%jobname%"

    return [
        "--stdout",
        join(oarsub_log_path, f"{base_name}.stdout"),
        "--stderr",
        join(oarsub_log_path, f"{base_name}.stderr"),
    ]



"""
Run code on the GriCAD cluster.
"""

import sys
import subprocess
from typing import Union, Optional
from os.path import join, isfile

from termcolor import colored
import click

from ...cli_utils import OrderedGroup, option_no_push
from ..cli_utils import option_script, argument_command, option_no_build
from .cli_utils import argument_oar_job_id, argument_oar_job_id_list
from ...utils import run_remote_command, run_remote_command_inria
from ...user_interaction import print_info, to_bold, print_error, print_warning
from ... import config
from ...file_transfer import (
    push_,
    option_push_force,
    pull_,
    argument_pull_remote_path,
    option_pull_force,
    clean_,
    argument_clean_directory,
    option_clean_force,
)
from ..singularity import get_binding_sub_command, build_singularity_container
from .oarsub import print_cluster_request_prop, add_standard_output_params


def _get_ssh_proxy_command(cluster_name: str = None) -> str:
    g_config: config.Gricad = config.Config().gricad
    if not cluster_name:
        cluster_name = g_config.prefered_cluster

    ssh_proxy_command: str = (
        f"ProxyCommand ssh -q {g_config.username}@access-gricad.univ-grenoble-alpes.fr"
        f" nc -w 60 {cluster_name} 22"
    )

    return ssh_proxy_command


def _run_remote_command_gricad(
    command: list[str],
    cluster_name: str = None,
    check: bool = False,
    force_tty_allocation: bool = False,
) -> Union[subprocess.Popen, int]:
    g_config: config.Gricad = config.Config().gricad

    if not cluster_name:
        cluster_name = g_config.prefered_cluster

    hostname_args: list[str] = [
        "-o",
        _get_ssh_proxy_command(cluster_name=cluster_name),
        f"{g_config.username}@{cluster_name}",
    ]

    return run_remote_command(
        command=command,
        hostname=hostname_args,
        pretty_hostname=cluster_name,
        check=check,
        force_tty_allocation=force_tty_allocation,
    )


# pylint: disable=unused-argument
def _callback_option_gricad_job_name(ctx, param, value) -> str:
    if not value:
        return config.Config().gricad.oarsub.job_name

    return value


_option_gricad_job_name = click.option(
    "-n",
    "--job-name",
    default="",
    callback=_callback_option_gricad_job_name,
    help="Custom job name",
)


# pylint: disable=unused-argument
def _callback_option_container(ctx, param, value) -> str:
    if value:
        return value

    return config.Config().gricad.singularity_image


_option_container: click.option = click.option(
    "-c",
    "--container",
    "container",
    type=str,
    callback=_callback_option_container,
    help="The container image to use (.sif)",
)


# pylint: disable=unused-argument
def _callback_option_gricad_gpu_model(ctx, param, value) -> str:
    if not value:
        value = config.Config().gricad.oarsub.gpu_model

    # Give more flexibility w.r.t the input: 'a100' --> 'A100'
    value = value.upper()

    if value not in ["A100", "V100", "T4"]:
        print_error(
            f"Invalid `gpu_model` '{value}'." " Supported values are: 'A100', 'V100' and 'T4'."
        )
        sys.exit(1)

    return value


_option_gricad_gpu_model = click.option(
    "-g",
    "--gpu-model",
    callback=_callback_option_gricad_gpu_model,
    help="GPU model ('A100', 'V100' or 'T4')",
)


@click.group(cls=OrderedGroup, invoke_without_command=True)
@option_script
@_option_gricad_gpu_model
@_option_gricad_job_name
@option_no_push
@click.pass_context
def gricad(
    ctx,
    script_name: str,
    job_name: str,
    gpu_model: str,
    no_push: bool,
) -> None:
    """
    Run a script on the cluster (see config for options).
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            cluster_script,
            script_name=script_name,
            job_name=job_name,
            gpu_model=gpu_model,
            no_push=no_push,
        )


#########
# Setup ############################################################################################
#########


@gricad.command()
def setup() -> None:
    """
    Set up remote project on the Gricad cluster.
    """
    _config: config.Config = config.Config()
    g_config: config.Gricad = _config.gricad

    print_info(
        text=f"Setting up gricad project: {_config.project_name}",
        bold=True,
    )

    paths_to_create: str = g_config.project_remote_path
    paths_to_create += (
        "/{"
        + ",".join(
            [
                _config.output_path,
                _config.notebooks_path,
                _config.oarsub_log_path,
            ]
        )
        + "}"
    )

    # Create the project directory on the remote and the output folder within it.
    _run_remote_command_gricad(
        command=["mkdir", "-p", paths_to_create],
        check=True,
    )
    _push_gricad()

    print_info("Setup complete !")


########
# Push #############################################################################################
########


def _push_gricad(force: bool = False):
    g_config: config.Gricad = config.Config().gricad

    cluster_name: str = g_config.prefered_cluster

    print_info("Pushing project folder to gricad", bold=True)

    rsync_destination: str = f"{g_config.username}@{cluster_name}:{g_config.project_remote_path}"
    rsync_e_opt: str = f'ssh -o "{_get_ssh_proxy_command(cluster_name=cluster_name)}"'

    push_(
        rsync_destination=rsync_destination,
        rsync_e_opt=rsync_e_opt,
        last_push_fn_suffix="_gricad",
        force=force,
    )


@gricad.command()
@option_push_force
def push(force: bool) -> None:
    """
    Sync the project folder to the gricad storage.

    Args:
        force (bool):   Run the sync command even if no local changes were detected.
    """
    _push_gricad(force=force)


########
# Pull #############################################################################################
########


@gricad.command()
@argument_pull_remote_path
@option_pull_force
def pull(remote_path: str, force: bool) -> None:
    """
    Sync the content of the provided remote directories to the equivalent folder to the local
    machine.

    Args:
        remote_path (str):  The remote folder to sync to the local computer.
        force (bool):       Do not ask for a confirmation before pulling.
                                Eventually conflicting local files might be overridden.
    """
    g_config: config.Gricad = config.Config().gricad

    rsync_e_opt: str = f'ssh -o "{_get_ssh_proxy_command()}"'

    # Source
    source_prefix: str = (
        f"{g_config.username}@{g_config.prefered_cluster}:{g_config.project_remote_path}"
    )

    pull_(
        remote_path=remote_path,
        source_prefix=source_prefix,
        rsync_e_opt=rsync_e_opt,
        force=bool,
    )


#########
# Clean ############################################################################################
#########


@gricad.command()
@option_clean_force
@argument_clean_directory
def clean(directory: str, force: bool) -> None:
    """
    Clean the content of a folder on the remote computer.

    Args:
        directory (str):        The directory of which content should be deleted.
        force (bool):           Do not ask for a confirmation before cleaning.
    """
    clean_cmd: Optional[list[str]] = clean_(
        directory=directory,
        remote_path=config.Config().gricad.project_remote_path,
        force=force,
    )
    if clean_cmd:
        _run_remote_command_gricad(clean_cmd, check=False)


###############
# Singularity ######################################################################################
###############
def _sync_container_from_inria() -> None:
    _config: config.Config = config.Config()
    singularity_config = _config.singularity
    def_file_name: str = singularity_config.def_file_name

    if not isfile(def_file_name):
        print_warning(f"Recipe file not found (`{def_file_name}`). Aborting")
        return

    # Compare gricad_last_synced_recipe.def to last_build_recipe.def
    last_built_recipe_path: str = ".remi/last_built_recipe.def"
    last_synced_recipe_path: str = ".remi/gricad_last_synced_recipe.def"

    last_built_recipe: str = ""
    if isfile(last_built_recipe_path):
        with open(last_built_recipe_path, "r") as last_recipe_built_file:
            last_built_recipe = last_recipe_built_file.read()

    last_synced_recipe: str = ""
    if isfile(last_synced_recipe_path):
        with open(last_synced_recipe_path, "r") as last_synced_recipe_file:
            last_synced_recipe = last_synced_recipe_file.read()

        # Check if something has changed
        if last_synced_recipe == last_built_recipe:
            print_info("Nothing has changed since last sync. Cancelling.")
            return

    # Else, something has changed: so we sync the image to gricad.

    rsync_e_opt: str = f'ssh -o "{_get_ssh_proxy_command()}"'
    inria_sif_name: str = singularity_config.output_sif_name
    path_to_sif_inria: str = join(
        _config.project_remote_path,
        inria_sif_name,
    )
    gricad_config: config.Gricad = _config.gricad
    path_to_sif_gricad: str = join(
        gricad_config.project_remote_path,
        gricad_config.singularity_image,
    )
    rsync_destination: str = (
        f"{gricad_config.username}@{gricad_config.prefered_cluster}:{path_to_sif_gricad}"
    )

    # Run singularity build on Inria remote
    sync_command: list[str] = [
        "rsync",
        "-avP",
        "-e",
        f"'{rsync_e_opt}'",
        path_to_sif_inria,
        rsync_destination,
    ]

    print_info(
        to_bold("Syncing container image ")
        + "`"
        + colored(inria_sif_name, "yellow")
        + "`"
        + to_bold(" to GRICAD: ")
        + "`"
        + colored(path_to_sif_gricad, "yellow")
        + "`"
    )

    return_code: int = run_remote_command_inria(
        command=sync_command,
        check=True,
        force_tty_allocation=True,
    )

    if return_code == 0:
        print_info("Sync succesfull.")
        # If the build command succeeded, save the current recipe as the 'last built recipe'
        subprocess.run(
            [
                "cp",
                "-f",
                last_built_recipe_path,
                last_synced_recipe_path,
            ],
            check=True,
        )
    else:
        print_error(f"Sync failed with return code {return_code}")


###########
# Cluster ##########################################################################################
###########


def _print_cluster_request(
    oarsub_config: config.GricadOarsub,
    job_name: str,
    cluster_name: str,
    gpu_model: str,
    container_sif: str,
) -> None:
    print_info(to_bold("Cluster request:"))
    print_cluster_request_prop(
        property_name="Gricad cluster",
        value=cluster_name,
    )
    print_cluster_request_prop(
        property_name="Job name",
        value=job_name,
    )
    print_cluster_request_prop(
        property_name="Apptainer image",
        value=container_sif,
    )
    print_cluster_request_prop(
        property_name="Number of nodes",
        value=oarsub_config.num_nodes,
    )
    if oarsub_config.num_cpus > 0:
        print_cluster_request_prop(
            property_name="Number of CPUs",
            value=oarsub_config.num_cpus,
        )
    if oarsub_config.num_cpu_cores > 0:
        print_cluster_request_prop(
            property_name="Number of CPU cores",
            value=oarsub_config.num_cpu_cores,
        )

    print_cluster_request_prop(
        property_name="Number of GPUs",
        value=oarsub_config.num_gpus,
    )
    if oarsub_config.num_gpus > 0 and gpu_model:
        print_cluster_request_prop(
            property_name="GPU model",
            value=gpu_model,
        )

    print_cluster_request_prop(
        property_name="Walltime",
        value=oarsub_config.walltime,
    )


def get_oarsub_cmd_prefix(
    job_name: str,
    cluster_name: str,
    gpu_model: str,
    container_sif: str,
) -> list[str]:
    # Get the oarsub config
    g_config: config.Gricad = config.Config().gricad
    oarsub_config: config.GricadOarsub = g_config.oarsub

    # Oarsub command
    oarsub_cmd: list[str] = ["oarsub"]

    # Name the job
    oarsub_cmd += ["--name", job_name]

    # Resource list
    resource_list: str = f"/nodes={oarsub_config.num_nodes}"
    if oarsub_config.num_cpus > 0:
        resource_list += f"/cpu={oarsub_config.num_cpus}"
    if oarsub_config.num_cpu_cores > 0:
        resource_list += f"/core={oarsub_config.num_cpu_cores}"

    if oarsub_config.num_gpus > 0:
        resource_list += f"/gpu={oarsub_config.num_gpus}"

    if oarsub_config.walltime != "":
        resource_list += f",walltime={oarsub_config.walltime}"

    oarsub_cmd += ["-l", resource_list]

    if oarsub_config.num_gpus > 0 and gpu_model:
        oarsub_cmd += ["-p", f"\"gpumodel='{gpu_model}'\""]

    oarsub_cmd += ["--project", g_config.project_name]

    _print_cluster_request(
        oarsub_config=oarsub_config,
        job_name=job_name,
        cluster_name=cluster_name,
        gpu_model=gpu_model,
        container_sif=container_sif,
    )

    return oarsub_cmd


def _cluster(
    sub_command: str,
    script_name: str = "",
    command: tuple[str] = "",
    gpu_model: str = "",
    job_name: str = "",
    container_sif: str = "",
    no_push: bool = False,
    no_build: bool = False,
) -> None:
    assert sub_command in ["script", "interactive", "command"]

    # Sync project.
    if not no_push:
        _push_gricad()
        print()

    _config: config.Config = config.Config()
    g_config: config.Gricad = _config.gricad

    cluster_name: str = g_config.prefered_cluster

    oarsub_cmd: list[str] = get_oarsub_cmd_prefix(
        job_name=job_name,
        cluster_name=cluster_name,
        gpu_model=gpu_model,
        container_sif=container_sif,
    )

    # Interactive
    if sub_command == "interactive":
        print_info("Connecting to cluster in interactive mode", bold=True)
        oarsub_cmd += ["--interactive"]

    # Script or command
    else:
        if not no_build:
            print()
            build_singularity_container(
                output_sif_name=container_sif,
            )
            _sync_container_from_inria()

        oarsub_cmd += add_standard_output_params(
            oarsub_log_path=join(
                g_config.project_remote_path,
                _config.oarsub_log_path,
            ),
            base_name=g_config.oarsub.log_file_name,
        )

        job_cmd: list[str] = ["cd", g_config.project_remote_path, "&&"]

        use_container: bool = g_config.use_container

        if use_container:
            job_cmd += ["/usr/local/bin/singularity", "exec"]

            job_cmd += get_binding_sub_command(
                project_remote_path=g_config.project_remote_path,
                env_variables=_config.singularity.env_variables,
            )

            # Enable GPU support and provide the path to the container image
            job_cmd += ["--nv", container_sif]

        if sub_command == "script":
            script_name_pretty: str = colored(script_name, "yellow")
            print_info(
                to_bold("Running script ")
                + colored(script_name_pretty, "yellow")
                + " on the cluster"
            )
            print()

            job_cmd += ["bash", script_name]

        elif sub_command == "command":
            command_merge: str = " ".join(command)
            command_merge_pretty: str = colored(command_merge, "yellow")
            print_info(
                to_bold("Running command ")
                + colored(command_merge_pretty, "yellow")
                + " on the cluster"
            )
            print()

            job_cmd += command

        if _config.verbose:
            job_cmd_pretty: str = colored(" ".join(job_cmd), "cyan")
            print_info(f"Command to run on the cluster: {job_cmd_pretty}")

        oarsub_cmd += ['"' + " ".join(job_cmd) + '"']

    _run_remote_command_gricad(
        command=oarsub_cmd,
        cluster_name=cluster_name,
        check=True,
        force_tty_allocation=True,
    )


@gricad.command("script")
@option_script
@_option_gricad_gpu_model
@_option_gricad_job_name
@_option_container
@option_no_push
@option_no_build
def cluster_script(
    script_name: str,
    gpu_model: str,
    job_name: str,
    container: str,
    no_push: bool,
    no_build: bool,
) -> None:
    _cluster(
        sub_command="script",
        script_name=script_name,
        gpu_model=gpu_model,
        job_name=job_name,
        container_sif=container,
        no_push=no_push,
        no_build=no_build,
    )


@gricad.command("command")
@argument_command
@_option_gricad_gpu_model
@_option_gricad_job_name
@_option_container
@option_no_push
@option_no_build
def cluster_command(
    command: list[str],
    job_name: str,
    gpu_model: str,
    container: str,
    no_push: bool,
    no_build: bool,
) -> None:
    _cluster(
        sub_command="command",
        command=command,
        gpu_model=gpu_model,
        job_name=job_name,
        container_sif=container,
        no_push=no_push,
        no_build=no_build,
    )


@gricad.command("interactive")
@_option_gricad_gpu_model
@_option_gricad_job_name
@option_no_push
def cluster_interactive(
    job_name: str,
    gpu_model: str,
    no_push: bool,
) -> None:
    _cluster(
        sub_command="interactive",
        job_name=job_name,
        gpu_model=gpu_model,
        no_push=no_push,
    )


@gricad.command("chandler")
def cluster_chandler() -> None:
    """
    Run `chandler` on the cluster to list compute nodes occupation (free/busy).
    """
    _run_remote_command_gricad(
        command="chandler",
        force_tty_allocation=True,
        check=True,
    )


@gricad.command("recap")
def cluster_recap() -> None:
    """
    Run `recap.py` on the cluster to list compute nodes information (CPU, GPU…).
    """
    _run_remote_command_gricad(
        command="recap.py",
        force_tty_allocation=True,
        check=True,
    )


@gricad.command("stat")
def cluster_stat() -> None:
    """
    Get some information about your running/planned jobs thanks to `oarstat`.
    """
    _run_remote_command_gricad(
        command=f"oarstat -u {config.Config().gricad.username}",
        force_tty_allocation=True,
        check=True,
    )


@gricad.command("connect")
@argument_oar_job_id
def cluster_connect(oar_job_id: int) -> None:
    """
    Connect to a job on the cluster.

    Arguments:
        OAR_JOB_ID: The id of the job you want to kill (It has to have been run by you)'
    """
    _run_remote_command_gricad(
        command=f"oarsub -C {oar_job_id}",
        force_tty_allocation=True,
        check=True,
    )


@gricad.command("kill")
@argument_oar_job_id_list
def gricad_kill(oar_job_ids: tuple[int]) -> None:
    """
    Kill a job on the gricad cluster.

    Arguments:
        OAR_JOB_IDS: The ids of the jobs you want to kill (they have to have been run by you)'
    """
    oar_job_ids_str: str = " ".join(str(job_id) for job_id in oar_job_ids)
    _run_remote_command_gricad(
        command=f"oardel {oar_job_ids_str}",
        check=True,
    )



"""
Utility functions for remi.
"""

import sys
import subprocess
from datetime import datetime, timedelta
from os.path import join
from socket import gethostname
from subprocess import Popen, DEVNULL
from typing import Union
from termcolor import colored

from .config import Config
from .user_interaction import to_bold, print_info, print_error


def run_local_cmd(
    command: list[str],
    check: bool = True,
    background: bool = False,
    redirect_output: bool = False,
) -> Union[subprocess.Popen, int]:
    """
    Run a command on the local computer.

    Args:
        command (list[str]):        The command to run.
        check (bool):               If True, raise an exception if the command fails.
        background (bool):          Run the command in the background.
        redirect_output (bool):     Redirects command standard and error outputs to
                                        `output/log.txt`.

    Returns:
        popen (subprocess.Popen):   If `background is True: the Popen object corresponding to the
                                        running command. (if background)
        return_code (int):          Else, the return code of the command.
    """
    config: Config = Config()
    if config.verbose:
        print_info(
            to_bold("Running command locally: ") + colored(" ".join(command), "blue"),
        )

    if redirect_output:
        log_file: str = join(config.output_path, "log.txt")
        command += [">>", log_file, "2>&1"]
        print_info(f"stdout and stderr will be logged to {log_file}")

    command = ["bash", "-c", " ".join(command)]

    # Run in background
    if background:
        print_info("Background mode enabled")
        # pylint: disable=consider-using-with
        return subprocess.Popen(
            args=command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            shell=False,
        )

    # If check is True, catch CalledProcessError exception and exit program with error.
    # Else, get the return_code and return it.
    try:
        return subprocess.run(command, check=check).returncode
    except subprocess.CalledProcessError:
        print_error(text="Local command failed. Exiting.")
        sys.exit(1)


def run_remote_command_inria(
    command: list[str],
    hostname: str = None,
    check: bool = False,
    background: bool = False,
    force_tty_allocation: bool = False,
    enable_x_forwarding: bool = False,
    redirect_output: bool = False,
    use_bastion: bool = None,
) -> Union[subprocess.Popen, int]:
    config: Config = Config()

    if not hostname:
        hostname = config.desktop.ip_adress

    if use_bastion is None:
        use_bastion = config.bastion.enable

    # If we are already connected to the remote, the command has to be run locally.
    if gethostname() in hostname:
        return run_local_cmd(
            command=command,
            check=check,
            background=background,
            redirect_output=redirect_output,
        )

    hostname_args: list[str] = []

    # Bastion ProxyJump (useless when working from an Inria computer)
    if not config.working_from_inria and use_bastion:
        hostname_args += ["-J", f"{config.username}@{config.bastion.hostname}"]

    hostname_args += [f"{config.username}@{hostname}"]

    return run_remote_command(
        command=command,
        hostname=hostname_args,
        pretty_hostname=hostname,
        check=check,
        background=background,
        force_tty_allocation=force_tty_allocation,
        enable_x_forwarding=enable_x_forwarding,
        redirect_output=redirect_output,
    )


def run_remote_command(
    command: list[str],
    hostname: str,
    pretty_hostname: str = None,
    check: bool = False,
    background: bool = False,
    force_tty_allocation: bool = False,
    enable_x_forwarding: bool = False,
    redirect_output: bool = False,
) -> Union[subprocess.Popen, int]:
    """
    Run a command on a remote computer.

    Args:
        command (list[str]):            A command to run remotely.
        hostname (str):                 The hostname of the computer to run the command on.
        pretty_hostname (str):          The hostname to display in the standard output.
        check (bool):                   If True, raise an exception if the command fails.
        background (bool):              Run the command in the background.
        force_tty_allocation (bool):    Force pseudo-terminal allocation.
        enable_x_forwarding (bool):     Enables X forwarding
        redirect_output (bool):         Redirects command standard and error outputs to
                                            `output/log.txt`.

    Returns:
        popen (subprocess.Popen):   If `background is True: the Popen object corresponding to the
                                        running command. (if background)
        return_code (int):          Else, the return code of the command.
    """
    config: Config = Config()

    cmd: list[str] = ["ssh"]

    if force_tty_allocation:
        cmd.append("-t")

    if enable_x_forwarding:
        cmd.append("-X")

    cmd += hostname if isinstance(hostname, list) else [hostname]

    if isinstance(command, str):
        command = [command]

    cmd += command

    if config.verbose:
        print_info(
            to_bold("Running command remotely: ") + colored(" ".join(command), "blue"),
        )
        if pretty_hostname is None:
            pretty_hostname = hostname[-1] if isinstance(hostname, list) else hostname

        print_info(
            to_bold("on remote host: ") + colored(str(pretty_hostname), "magenta"),
        )

    if redirect_output:
        log_file: str = join(config.output_path, "log.txt")
        cmd += [">>", log_file, "2>&1"]
        print_info(f"stdout and stderr will be logged to {log_file}")
        print_info("Use `remi pull-output` to sync them to your local machine.")

    # Run in background
    if background:
        print_info("Background mode enabled")
        # pylint: disable=consider-using-with
        return subprocess.Popen(
            args=cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            shell=False,
        )

    # Check the time the command takes to run
    start_time: float = datetime.now()

    return_code: int

    # Run directly (not in background)
    # If check is True, catch CalledProcessError exception and exit program with error.
    # Else, get the return_code and return it.
    try:
        return_code = subprocess.run(cmd, check=check).returncode
    except subprocess.CalledProcessError:
        print_error(
            text="Remote command failed. Exiting.",
            newline=True,
        )
        sys.exit(1)

    duration: timedelta = datetime.now() - start_time
    print_info(
        f"Command ran succesfully. It took: {str(duration).split('.')[0]}.",
        newline=True,
    )

    return return_code


def create_remote_directory(
    path: str,
    hostname: str = None,
) -> None:
    """
    Create a directory on the remote desktop if it does not exist yet.

    Args:
        path (str):         The absolute path to the remote directory.
        hostname (str):     The hostname of the remote computer.
    """
    run_remote_command_inria(
        command=["mkdir", "-p", path],
        hostname=hostname,
        check=True,
    )


def run_ssh_tunnel(
    hostname: str,
    port: int,
) -> Popen:
    config: Config = Config()

    print_info(f"Running SSH port forwarding for port {port}")
    print_info(to_bold("on host: ") + str(hostname))
    ssh_tunnel_cmd: list[str] = ["ssh", "-N"]

    ssh_tunnel_cmd += ["-L", f"{port}:localhost:{port}"]

    # Bastion ProxyJump
    if config.bastion.enable:
        ssh_tunnel_cmd += ["-J", f"{config.username}@{config.bastion.hostname}"]

    ssh_tunnel_cmd.append(f"{config.username}@{hostname}")

    print_info(to_bold("SSH tunnel command: ") + " ".join(ssh_tunnel_cmd))

    # pylint: disable=consider-using-with
    return Popen(
        args=ssh_tunnel_cmd,
        stdout=DEVNULL,
        stderr=DEVNULL,
    )



"""
Functions specific to the Inria cluster.
"""

from os.path import join
import subprocess
from typing import Union

import click
from termcolor import colored

from ...config import Config
from ...user_interaction import to_bold, print_info
from ...file_transfer import push_inria
from ...utils import run_remote_command_inria
from ...cli_utils import OrderedGroup, option_no_push, option_xforwarding
from ..cli_utils import option_script, option_no_build, argument_command, option_container
from .cli_utils import argument_oar_job_id, argument_oar_job_id_list

from ..singularity import get_binding_sub_command, build_singularity_container
from .oarsub import get_oarsub_cmd_prefix, add_standard_output_params


def _run_remote_command_cluster(
    command: list[str],
    enable_x_forwarding: bool = False,
) -> Union[subprocess.Popen, int]:
    return run_remote_command_inria(
        command=command,
        hostname="access2-cp.inrialpes.fr",
        check=True,
        force_tty_allocation=True,
        enable_x_forwarding=enable_x_forwarding,
        use_bastion=True,
    )


def _cluster(
    sub_command: str,
    script_name: str = "",
    command: tuple[str] = "",
    host_id: str = "",
    job_name: str = "",
    container_sif: str = "",
    enable_x_forwarding: bool = False,
    no_build: bool = False,
    no_push: bool = False,
) -> None:
    assert sub_command in ["script", "interactive", "command"]

    # Sync project.
    if not no_push:
        push_inria()
        print()

    config: Config = Config()

    oarsub_cmd: list[str] = get_oarsub_cmd_prefix(
        job_name=job_name,
        host_id=host_id,
    )

    # Interactive
    if sub_command == "interactive":
        print_info("Connecting to cluster in interactive mode", bold=True)
        oarsub_cmd += ["--interactive"]

    # TODO

    # Script or command
    else:
        # X11 xforwarding makes no sense when not in interactive mode.
        enable_x_forwarding = False

        if not no_build:
            print()
            build_singularity_container(
                output_sif_name=container_sif,
            )
            print()

        oarsub_cmd += add_standard_output_params(
            oarsub_log_path=join(
                config.project_remote_path,
                config.oarsub_log_path,
            ),
            base_name=config.oarsub.log_file_name,
        )

        job_cmd: list[str] = ["cd", config.project_remote_path, "&&", "singularity", "exec"]

        # Path bindings between host system and singularity container
        job_cmd += get_binding_sub_command(
            project_remote_path=config.project_remote_path,
            bindings=config.singularity.bindings,
            env_variables=config.singularity.env_variables,
            no_home=config.singularity.no_home,
        )

        if config.singularity.bind_beegfs:
            job_cmd += ["--bind", f"/services/scratch/perception/{config.username}/:/beegfs/"]

        # Export the PYTHONPATH environment variable so as to use the project python module
        # job_cmd += ['--env', 'PYTHONPATH=.']

        if config.oarsub.num_gpus > 0:
            job_cmd.append("--nv")

        job_cmd.append(container_sif)

        if sub_command == "script":
            script_name_pretty: str = colored(script_name, "yellow")
            print_info(
                to_bold("Running script ")
                + colored(script_name_pretty, "yellow")
                + " on the cluster"
            )
            print()

            job_cmd += ["bash", script_name]

        elif sub_command == "command":
            command_merge: str = " ".join(command)
            command_merge_pretty: str = colored(command_merge, "yellow")
            print_info(
                to_bold("Running command ")
                + colored(command_merge_pretty, "yellow")
                + " on the cluster"
            )
            print()

            job_cmd += command

        if config.verbose:
            job_cmd_pretty: str = colored(" ".join(job_cmd), "cyan")
            print_info(f"Command to run on the cluster: {job_cmd_pretty}")

        oarsub_cmd += ['"' + " ".join(job_cmd) + '"']

    _run_remote_command_cluster(
        command=oarsub_cmd,
        enable_x_forwarding=enable_x_forwarding,
    )


#######
# CLI #
#######
# pylint: disable=unused-argument
def _callback_option_cluster_job_name(ctx, param, value) -> str:
    if not value:
        return Config().oarsub.job_name

    return value


_option_cluster_job_name = click.option(
    "-n",
    "--job-name",
    default="",
    callback=_callback_option_cluster_job_name,
    help="Custom job name",
)


# pylint: disable=unused-argument
def _callback_option_host_id(ctx, param, value) -> str:
    if not value:
        return Config().oarsub.host_id

    return value


_option_host_id = click.option(
    "-h",
    "--host",
    "host_id",
    default="",
    callback=_callback_option_host_id,
    type=str,
    help="A specific host id.",
)


@click.group(cls=OrderedGroup, invoke_without_command=True)
@option_script
@_option_host_id
@_option_cluster_job_name
@option_container
@option_no_push
@option_no_build
@click.pass_context
def cluster(
    ctx,
    script_name: str,
    job_name: str,
    host_id: str,
    container: str,
    no_push: bool,
    no_build: bool,
) -> None:
    """
    Run a script on the cluster (see config for options).
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            cluster_script,
            script_name=script_name,
            job_name=job_name,
            host_id=host_id,
            container=container,
            no_push=no_push,
            no_build=no_build,
        )


@cluster.command("script")
@option_script
@_option_host_id
@_option_cluster_job_name
@option_container
@option_no_push
@option_no_build
def cluster_script(
    script_name: str,
    host_id: str,
    job_name: str,
    container: str,
    no_push: bool,
    no_build,
) -> None:
    _cluster(
        sub_command="script",
        script_name=script_name,
        host_id=host_id,
        job_name=job_name,
        container_sif=container,
        no_push=no_push,
        no_build=no_build,
    )


@cluster.command("command")
@argument_command
@_option_host_id
@_option_cluster_job_name
@option_container
@option_no_push
@option_no_build
def cluster_command(
    command: list[str],
    host_id: str,
    job_name: str,
    container: str,
    no_push: bool,
    no_build: bool,
) -> None:
    _cluster(
        sub_command="command",
        command=command,
        host_id=host_id,
        job_name=job_name,
        container_sif=container,
        no_push=no_push,
        no_build=no_build,
    )


@cluster.command("interactive")
@_option_host_id
@_option_cluster_job_name
@option_xforwarding
@option_no_push
def cluster_interactive(
    host_id: str,
    job_name: str,
    x_forwarding: bool,
    no_push: bool,
) -> None:
    _cluster(
        sub_command="interactive",
        host_id=host_id,
        job_name=job_name,
        enable_x_forwarding=x_forwarding,
        no_push=no_push,
    )


@cluster.command("stat")
@click.option(
    "-t",
    "--team",
    default=False,
    is_flag=True,
    help="Query the current jobs for the RobotLearn team members",
)
def cluster_stat(
    team: bool,
) -> None:
    """
    Get the state of your jobs.
    """
    command: list[str] = ["oarstat"]

    if team:
        command += [
            "|",
            "grep",
            "'"
            + "\\|".join(
                (
                    "aauterna",
                    "aballou",
                    "creinke",
                    "galepage",
                    "lairale",
                    "visanche",
                    "wguo",
                    "xilin",
                    "xbie",
                    "yixu",
                )
            )
            + "'",
        ]
    else:
        command += ["-u", Config().username]

    _run_remote_command_cluster(
        command=command,
    )


@cluster.command("connect")
@argument_oar_job_id
def cluster_connect(oar_job_id: int) -> None:
    """
    Connect to a job on the cluster.

    Arguments:
        OAR_JOB_ID: The id of the job you want to kill (It has to have been run by you)'
    """
    _run_remote_command_cluster(
        command=f"oarsub -C {oar_job_id}",
    )


@cluster.command("kill")
@argument_oar_job_id_list
def cluster_kill(oar_job_ids: tuple[int]) -> None:
    """
    Kill a job on the cluster.

    Arguments:
        OAR_JOB_IDS: The ids of the jobs you want to kill (they have to have been run by you)'
    """
    oar_job_ids_str: str = " ".join(str(job_id) for job_id in oar_job_ids)
    _run_remote_command_cluster(
        command=f"oardel {oar_job_ids_str}",
    )



"""
Common options for remote compute commands
"""

import sys
from os.path import isfile
import click

from ..user_interaction import print_error
from ..config import Config


# pylint: disable=unused-argument
def _callback_option_script(ctx, param, value) -> str:
    # Make sure the script is a valid file
    if ctx.invoked_subcommand not in (None, "script"):
        if not isfile(value):
            print_error(f"Script `{value}` does not exist.")
            sys.exit(1)

    return value


# pylint: disable=unused-argument
def _callback_argument_command(ctx, param, value) -> str:
    if len(value) == 0:
        print_error("No command provided. Exiting")
        sys.exit(1)

    return value


# pylint: disable=unused-argument
def _callback_option_container(ctx, param, value) -> str:
    if value:
        return value

    return Config().singularity.output_sif_name


option_interactive = click.option(
    "-i",
    "--interactive",
    is_flag=True,
    help="Open an interactive session instead of running the script",
)

option_script = click.option(
    "-s",
    "--script",
    "script_name",
    default="script.sh",
    type=str,
    callback=_callback_option_script,
    help="Path to a bash script to run",
)

option_background = click.option(
    "-b",
    "--background",
    default=False,
    is_flag=True,
    help="Run the script in the background",
)

option_no_build = click.option(
    "--no-build",
    default=False,
    is_flag=True,
    help="Prevent trying to build the singularity container.",
)

argument_command = click.argument(
    "command",
    nargs=-1,
    type=str,
    callback=_callback_argument_command,
)


option_container: click.option = click.option(
    "-c",
    "--container",
    "container",
    type=str,
    callback=_callback_option_container,
    help="The container image to use (.sif)",
)



"""
Commands for setting up remi for a project.
"""

import click

from .config import Config
from .config.defaults import create_default_project
from .user_interaction import print_info, prompt_user_yes_no
from .utils import create_remote_directory, run_local_cmd
from .cli_utils import option_hostname
from .file_transfer import push_inria


@click.command()
def init() -> None:
    """
    Initialize the project in the current working directory. Generate the configuration files.
    """
    print_info(
        text="Creating a brand new remi project.",
        bold=True,
    )
    create_default_project()


@click.command()
@option_hostname
def setup(hostname: str) -> None:
    """
    Set up remote project location.

    Args:
        hostname (str):     The hostname of the remote computer.
    """
    config: Config = Config()

    print_info(
        text=f"Setting up remote for project: {config.project_name}",
        bold=True,
    )

    paths_to_create: str = config.project_remote_path
    paths_to_create += (
        "/{"
        + ",".join(
            [
                config.output_path,
                config.notebooks_path,
                config.oarsub_log_path,
            ]
        )
        + "}"
    )

    # Create the project directory on the remote and the output folder within it.
    create_remote_directory(
        path=paths_to_create,
        hostname=hostname,
    )
    push_inria()

    print_info("Setup complete !")


@click.command()
def update() -> None:
    """
    Update remi to the latest version
    """
    if prompt_user_yes_no("Do you wish to update 'remi' to the latest version ?"):
        run_local_cmd(
            command=[
                "pip",
                "install",
                "-U",
                "remote-inria",
            ],
            check=True,
        )
    else:
        print_info("All right the, aborting.")



"""
Functions to interact with singularity.
"""

import sys
from typing import Optional
import subprocess
from os.path import isfile
from termcolor import colored

import click

from ..cli_utils import option_force
from ..user_interaction import to_bold, print_info, print_warning, print_error
from ..utils import run_remote_command_inria
from ..file_transfer import push_inria
from ..config import Config


def get_binding_sub_command(
    project_remote_path: str,
    bindings: dict[str, Optional[str]] = None,
    env_variables: dict[str, str | int] = None,
    no_home: bool = False,
) -> list[str]:
    """
    Path bindings between host system and singularity container
    """
    binding_sub_cmd: list[str] = []

    if bindings is not None:
        if project_remote_path not in bindings:
            binding_sub_cmd += ["--bind", project_remote_path]
        for host_path, container_path in bindings.items():
            binding_sub_cmd += ["--bind", host_path]
            if container_path:
                binding_sub_cmd[-1] += ":" + container_path
    else:
        binding_sub_cmd += ["--bind", project_remote_path]

    if env_variables:
        for var_name, var_value in env_variables.items():
            var_value_str: str = ""
            match var_value:
                case int():
                    var_value_str = str(var_value)
                case str():
                    var_value_str = f'"{var_value}"'
                case None:
                    var_value_str = ""
                case _:
                    print_error(
                        f"Value for environment variable `${var_name}` has an unexpected type."
                        f"\nType: {type(var_value)}.\nValue: {str(var_value)}"
                    )
                    sys.exit(1)

            binding_sub_cmd += ["--env", f"{var_name}={var_value_str}"]

    if no_home:
        binding_sub_cmd.append("--no-home")

    return binding_sub_cmd


def build_singularity_container(
    output_sif_name: str = "",
    force: bool = False,
) -> None:
    singularity_config = Config().singularity

    def_file_name: str = singularity_config.def_file_name

    if output_sif_name == "":
        output_sif_name = singularity_config.output_sif_name

    # If not using the default sif, don't try to build
    elif output_sif_name != singularity_config.output_sif_name:
        return

    if not isfile(def_file_name):
        print_warning(f"Recipe file not found (`{def_file_name}`). Aborting")
        return

    # Smart build
    # The 'smart build' mechanism stores the last built recipe file at each build
    # under `.remi/last_built_recipe.def`.
    # Then we are able to check if something has changed locally since last build and run the
    # singularity command only if necessary (if `container.def` differs from
    # `.remi/last_built_recipe.def`).
    # This avoids opening a useless ssh tunnel with the server (which can sometimes last a few
    # seconds) and asking a confirmation to the user.
    last_built_recipe_path: str = ".remi/last_built_recipe.def"

    last_built_recipe: str = ""
    if isfile(last_built_recipe_path):
        with open(last_built_recipe_path, "r") as last_recipe_built_file:
            last_built_recipe = last_recipe_built_file.read()

    with open(def_file_name, "r") as current_recipe_file:
        current_recipe: str = current_recipe_file.read()

    # Check if something has changed
    if current_recipe == last_built_recipe and not force:
        print_info("Nothing has changed since last build. Cancelling build.")
        return

    # Else, something has changed: so we build the image remotely.

    # Run singularity build on Inria remote
    build_command: list[str] = [
        "cd",
        Config().project_remote_path,
        "&&",
        "sudo",
        "singularity",
        "build",
    ]

    print_info(
        to_bold("Building container image ")
        + "`"
        + colored(output_sif_name, "yellow")
        + "`"
        + to_bold(" from recipe ")
        + "`"
        + colored(def_file_name, "yellow")
        + "`"
    )

    if force:
        build_command += ["--force"]

    build_command += [output_sif_name, def_file_name]

    return_code: int = run_remote_command_inria(
        command=build_command,
        check=False,
        force_tty_allocation=True,
    )

    if return_code == 0:
        print_info("Build succesfull.")
        # If the build command succeeded, save the current recipe as the 'last built recipe'
        subprocess.run(
            [
                "cp",
                "-f",
                def_file_name,
                last_built_recipe_path,
            ],
            check=True,
        )
    else:
        print_info("You canceled the build remotely")


@click.command("build-container")
@option_force(help_message="Build the container even if the sif image already exists.")
def build_container(force: bool) -> None:
    """
    Build the singularity container on the remote desktop.
    """
    # First, sync project with remote
    push_inria()

    build_singularity_container(force=force)



"""
Common options for remi commands
"""

from collections import OrderedDict
import click
import re

from .config import Config

# Common options, arguments and their eventual callbacks


# pylint: disable=unused-argument
def _callback_option_hostname(ctx, param, value) -> str:
    config: Config = Config()
    # If no hostname was provided, use the one from the config file.
    if value == "":
        return config.desktop.ip_adress

    if re.match(r"gpu[0-9]$", value):
        value += "-perception"

    # Make sure the provided hostname is a valid Inria hostname
    if not value.endswith(".inrialpes.fr"):
        value += ".inrialpes.fr"

    return value


option_hostname: click.option = click.option(
    "-h",
    "--hostname",
    default="",
    type=str,
    callback=_callback_option_hostname,
    help="Name of an Inria computer",
)
option_no_push: click.option = click.option(
    "--no-push",
    default=False,
    is_flag=True,
    help="Prevent syncing the project files.",
)

option_xforwarding = click.option(
    "-X",
    "--xforwarding",
    "x_forwarding",
    default=False,
    is_flag=True,
    help="Enables X forwarding in SSH.",
)


def option_force(
    help_message: str,
    default: bool = False,
):
    return click.option(
        "-f",
        "--force",
        default=default,
        is_flag=True,
        help=help_message,
    )


class OrderedGroup(click.Group):
    """
    Override of the click.Group class in order to preserve custom ordering of commands in `--help`.
    """

    def __init__(
        self,
        name=None,
        commands=None,
        **attrs,
    ) -> None:
        super().__init__(name, commands, **attrs)
        #: the registered subcommands by their exported names.
        self.commands = commands or OrderedDict()

    def list_commands(self, ctx):
        """
        Override of the `list_commands` method.
        """
        return self.commands



"""
Functions to interact with the user through the command line.
"""

from typing import Any

from termcolor import colored


def to_bold(text: str):
    return colored(text, attrs=["bold"])


def print_info(
    text: str,
    newline: bool = False,
    bold: bool = False,
    end: str = "\n",
) -> None:
    prefix: str = ""
    if newline:
        prefix += "\n"
    prefix += colored("[INFO]", "green", attrs=["bold"])

    if bold:
        text = to_bold(text)

    print(prefix, text, end=end)


def print_warning(
    text: str,
    newline: bool = False,
    bold: bool = False,
    end: str = "\n",
) -> None:
    prefix: str = ""
    if newline:
        prefix += "\n"
    prefix += colored("[WARNING]", "yellow", attrs=["bold"])

    if bold:
        text = to_bold(text)

    print(prefix, text, end=end)


def print_error(
    text: str,
    newline: bool = False,
    bold: bool = False,
    end: str = "\n",
) -> None:
    prefix: str = ""
    if newline:
        prefix += "\n"
    prefix += colored("[ERROR]", "red", attrs=["bold"])

    if bold:
        text = to_bold(text)

    print(prefix, text, end=end)


def prompt_user_yes_no(question: str) -> bool:
    answer: str = ""
    while answer not in ["y", "n"]:
        print_warning(question, end="")
        answer = input(" [y/n] ").lower()

    return answer == "y"


def prompt_user_value(
    question: str,
    default_value: Any = None,
) -> Any:
    question: str = question

    if default_value is not None:
        question += f" [{colored(default_value, 'yellow')}]"
    question += ": "

    return input(question).strip() or default_value



"""
Run code on a remote desktop computer.
"""

from socket import gethostname
import click
import re
import sys

from ..cli_utils import OrderedGroup, option_hostname, option_xforwarding, option_no_push
from .cli_utils import (
    option_script,
    option_background,
    option_no_build,
    argument_command,
    option_container,
)

from ..config import Config
from ..user_interaction import to_bold, print_info, print_warning, colored, print_error
from ..utils import run_remote_command_inria
from ..file_transfer import push_inria
from .singularity import build_singularity_container, get_binding_sub_command


def _desktop(
    sub_command: str,
    hostname: str,
    script_name: str = "",
    background: bool = False,
    session_name: str = "",
    container_sif: str = "",
    attach: bool = False,
    command: tuple[str] = "",
    gpus: str = "",
    enable_x_forwarding: bool = False,
    no_push: bool = False,
    no_build: bool = False,
) -> None:
    assert sub_command in ["script", "interactive", "command"]

    # Sync project.
    if not no_push:
        push_inria()

    config: Config = Config()

    cmd: list[str] = ["cd", config.project_remote_path, "&&"]

    singularity_cmd: list[str] = []

    use_container: bool = config.desktop.use_container

    if use_container:
        if not no_build:
            build_singularity_container(
                output_sif_name=container_sif,
            )
            print()

        singularity_cmd = ["singularity"]
        if sub_command == "interactive":
            singularity_cmd.append("shell")
        else:
            singularity_cmd.append("exec")

        singularity_cmd += get_binding_sub_command(
            project_remote_path=config.project_remote_path,
            bindings=config.singularity.bindings,
            env_variables=config.singularity.env_variables,
            no_home=config.singularity.no_home,
        )

        if config.singularity.bind_beegfs:
            singularity_cmd += ["--bind", f"/mnt/beegfs/perception/{config.username}/:/beegfs/"]

        # Export the PYTHONPATH environment variable so as to use the project python module
        singularity_cmd += ["--env", "PYTHONPATH=."]

        if gpus:
            singularity_cmd += ["--env", f"CUDA_VISIBLE_DEVICES={gpus}"]

        # Enable GPU support and provide the path to the container image
        singularity_cmd += ["--nv", container_sif]

    # Interactive
    if sub_command == "interactive":
        if gethostname() in hostname:
            print_info(
                "Running an interactive session on this workstation.",
                bold=True,
            )

        else:
            print_info(
                f"Connecting to remote desktop ({hostname}) in interactive mode.",
                bold=True,
            )

        if use_container:
            cmd += singularity_cmd

        # No container
        else:
            cmd += ["exec", "bash"]

    # Script or command
    else:
        cmd_suffix: str = ""

        # Script
        if sub_command == "script":
            print_info(
                to_bold("Running script ")
                + colored(script_name, "yellow")
                + " on remote desktop "
                + colored(hostname, "magenta")
            )
            print()

            cmd_suffix += f"bash {script_name}"

        # Command
        else:
            assert len(command) > 0

            command_merge: str = " ".join(command)
            print_info(
                to_bold("Running command ")
                + colored(command_merge, "yellow")
                + " on remote desktop "
                + colored(hostname, "magenta")
            )
            print()

            cmd_suffix += command_merge

        # Use `screen` to run a command in the background.
        if background:
            if config.desktop.background.backend == "screen":
                cmd += ["screen", "-dmS"]
            elif config.desktop.background.backend == "tmux":
                cmd += ["tmux", "new-session", "-d"]

                # Set the session working directory to the project remote path
                cmd += ["-c", config.project_remote_path]

                # Name the session
                cmd += ["-s"]

            cmd += [session_name, "bash", "-c"]

            # If singularity + container, add singularity prefix in the cmd_suffix.
            if use_container:
                cmd_suffix = " ".join(singularity_cmd) + " " + cmd_suffix

            # Keep session alive
            if config.desktop.background.keep_session_alive:
                print_warning(
                    f"The `{config.desktop.background.backend}` session will be kept alive once the"
                    " job command will have exited.\n"
                    "You can run `remi desktop attach-session` to re-attach to the session."
                )
                cmd_suffix += "; exec bash"

            # Keep the session running only if the command fails
            else:
                cmd_suffix += " || exec bash"

            cmd_suffix = '"' + cmd_suffix + '"'

        # If not in background, the singularity cmd is appended as a list of arguments.
        elif use_container:
            cmd += singularity_cmd

        cmd.append(cmd_suffix)

    run_remote_command_inria(
        command=cmd,
        hostname=hostname,
        check=True,
        force_tty_allocation=True,
        enable_x_forwarding=enable_x_forwarding,
    )

    if background and attach:
        _attach_session(
            hostname=hostname,
            session_name=session_name,
        )


def _attach_session(
    hostname: str,
    session_name: str,
) -> None:
    background_backend: str = Config().desktop.background.backend

    cmd: list[str]
    if background_backend == "screen":
        cmd = ["screen", "-r", session_name]
    elif background_backend == "tmux":
        cmd = ["tmux", "attach-session", "-t", session_name]

    run_remote_command_inria(
        cmd,
        hostname=hostname,
        check=True,
        force_tty_allocation=True,
    )


#######
# CLI #
#######
_option_attach: click.option = click.option(
    "-a",
    "--attach",
    default=False,
    is_flag=True,
    help="Automatically attach to screen/tmux session (Only useful when `background` is enabled.",
)


# pylint: disable=unused-argument
def _callback_option_session_name(ctx, param, value) -> str:
    # Make sure the script is a valid file
    if value:
        return value

    return Config().project_name


_option_session_name: click.option = click.option(
    "-n",
    "--name",
    "session_name",
    type=str,
    callback=_callback_option_session_name,
    help="Manually name the screen/tmux session",
)


# pylint: disable=unused-argument
def _callback_option_gpus(ctx, param, value) -> str:
    if (value is not None) and (not re.fullmatch(r"[0-9](,[0-9])*", value)):
        print_error(
            f"Provided value (`{value}`) for the -g/--gpus option is invalid."
            "\nIt should be one or several GPU ids."
            "\nExamples: `0`, `0,1`, `1,3,4`"
        )
        sys.exit(1)

    return value


_option_gpus: click.option = click.option(
    "-g",
    "--gpus",
    "gpus",
    type=str,
    callback=_callback_option_gpus,
    help="The value for `$CUDA_VISIBLE_DEVICES` (e.g.: '1,2')",
)


@click.group(cls=OrderedGroup, invoke_without_command=True)
@click.pass_context
@option_script
@option_hostname
@option_background
@_option_session_name
@option_container
@_option_attach
@_option_gpus
@option_xforwarding
@option_no_push
@option_no_build
def desktop(
    ctx,
    script_name: str,
    background: bool,
    session_name: str,
    container: str,
    attach: bool,
    hostname: str,
    gpus: str,
    x_forwarding: bool,
    no_push: bool,
    no_build: bool,
) -> None:
    """
    Run a script on the remote workstation.
    """
    if ctx.invoked_subcommand is None:
        ctx.invoke(
            desktop_script,
            script_name=script_name,
            hostname=hostname,
            background=background,
            session_name=session_name,
            container=container,
            attach=attach,
            gpus=gpus,
            x_forwarding=x_forwarding,
            no_push=no_push,
            no_build=no_build,
        )


@desktop.command("script")
@option_script
@option_hostname
@option_background
@_option_session_name
@option_container
@_option_attach
@_option_gpus
@option_xforwarding
@option_no_push
@option_no_build
def desktop_script(
    script_name: str,
    hostname: str,
    background: bool,
    session_name: str,
    container: str,
    attach: bool,
    gpus: str,
    x_forwarding: bool,
    no_push: bool,
    no_build,
) -> None:
    _desktop(
        sub_command="script",
        script_name=script_name,
        hostname=hostname,
        background=background,
        session_name=session_name,
        container_sif=container,
        attach=attach,
        gpus=gpus,
        enable_x_forwarding=x_forwarding,
        no_push=no_push,
        no_build=no_build,
    )


@desktop.command("command")
@option_hostname
@option_background
@_option_session_name
@option_container
@_option_attach
@_option_gpus
@option_xforwarding
@option_no_push
@option_no_build
@argument_command
def desktop_command(
    command: list[str],
    background: bool,
    session_name: str,
    container: str,
    attach: bool,
    hostname: str,
    gpus: str,
    x_forwarding: bool,
    no_push: bool,
    no_build: bool,
) -> None:
    _desktop(
        sub_command="command",
        command=command,
        background=background,
        session_name=session_name,
        container_sif=container,
        attach=attach,
        hostname=hostname,
        gpus=gpus,
        enable_x_forwarding=x_forwarding,
        no_push=no_push,
        no_build=no_build,
    )


@desktop.command("interactive")
@option_hostname
@option_container
@_option_gpus
@option_xforwarding
@option_no_push
@option_no_build
def desktop_interactive(
    hostname: str,
    container: str,
    gpus: str,
    x_forwarding: bool,
    no_push: bool,
    no_build: bool,
) -> None:
    _desktop(
        sub_command="interactive",
        hostname=hostname,
        container_sif=container,
        gpus=gpus,
        enable_x_forwarding=x_forwarding,
        no_push=no_push,
        no_build=no_build,
    )


argument_session_name: click.argument = click.argument(
    "session_name",
    default="",
    type=str,
    callback=_callback_option_session_name,
    nargs=1,
)


@desktop.command("attach-session")
@argument_session_name
@option_hostname
def desktop_attach_session(
    hostname: str,
    session_name: str,
) -> None:
    _attach_session(
        hostname=hostname,
        session_name=session_name,
    )



"""
Configuration management for remi.
"""

import sys
from os.path import join, isfile, basename, abspath
from typing import Any
from abc import ABC
from dataclasses import dataclass
from socket import gethostname
import yaml

from ..user_interaction import print_error

_DOT_REMI_PATH: str = ".remi/"
_CONFIG_FILE_NAME: str = "config.yaml"

_CONFIG = None


class Singleton(type):
    """
    Implementation of the singleton pattern.
    """

    _instances: dict = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]


@dataclass
class Background:
    backend: str
    keep_session_alive: bool

    @classmethod
    def from_dict(cls, args_dict: dict[str, Any]) -> "Background":
        return cls(
            backend=args_dict.get("backend", "screen"),
            keep_session_alive=args_dict.get("keep_session_alive", False),
        )


@dataclass
class Desktop:
    hostname: str
    ip_adress: str
    use_container: bool
    background: Background

    @classmethod
    def from_dict(cls, args_dict: dict[str, Any]) -> "Desktop":
        background_dict: dict[str, Any] = args_dict.get("background", {})
        background: Background = Background.from_dict(args_dict=background_dict)
        assert background.backend in (
            "screen",
            "tmux",
        ), "`background.backend` should be one of `screen` or `tmux`."

        return cls(
            hostname=args_dict.get("hostname"),
            ip_adress=args_dict.get("ip_adress"),
            use_container=args_dict.get("use_container", True),
            background=background,
        )


@dataclass
class Bastion:
    enable: bool
    hostname: str
    username: str

    @classmethod
    def from_dict(cls, args_dict: dict[str, Any]) -> "Bastion":
        return cls(
            enable=args_dict.get("enable", True),
            hostname=args_dict.get("hostname"),
            username=args_dict.get("username"),
        )


@dataclass
class Oarsub:
    _default_walltime: str = "72:00:00"

    job_name: str = ""
    num_hosts: int = 1
    num_cpu_cores: int = 0
    num_gpus: int = 0
    walltime: str = _default_walltime
    cluster_name: str = ""
    host_id: str = ""
    custom_property_query: str = ""
    besteffort: bool = True
    idempotent: bool = False
    log_file_name: str = ""

    @classmethod
    def from_dict(cls, args_dict: dict[str, Any], default_job_name: str) -> "Oarsub":
        return cls(
            job_name=args_dict.get("job_name", default_job_name),
            num_hosts=args_dict.get("num_hosts", 1),
            num_cpu_cores=args_dict.get("num_cpu_cores", 0),
            num_gpus=args_dict.get("num_gpus", 1),
            walltime=args_dict.get("walltime", cls._default_walltime),
            cluster_name=args_dict.get("cluster_name", "perception"),
            host_id=args_dict.get("host_id", ""),
            custom_property_query=args_dict.get("custom_property_query", ""),
            besteffort=args_dict.get("besteffort", True),
            idempotent=args_dict.get("idempotent", False),
            log_file_name=args_dict.get("log_file_name", ""),
        )


@dataclass
class Singularity:
    def_file_name: str
    output_sif_name: str
    bindings: dict[str, str]
    bind_beegfs: bool
    env_variables: dict[str, str | int]
    no_home: bool

    @classmethod
    def from_dict(
        cls,
        args_dict: dict[str, Any],
    ) -> "Singularity":
        return cls(
            output_sif_name=args_dict.get("output_sif_name", "container.sif"),
            def_file_name=args_dict.get("def_file_name", "container.def"),
            bindings=args_dict.get("bindings", {}),
            bind_beegfs=args_dict.get("bind_beegfs", False),
            env_variables=args_dict.get("env_variables", {}),
            no_home=args_dict.get("no_home", False),
        )


@dataclass
class GricadOarsub(Oarsub):
    _default_walltime: str = "48:00:00"

    num_nodes: int = 1
    num_cpus: int = 0
    gpu_model: str = ""
    log_file_name: str = ""

    @classmethod
    def from_dict(
        cls,
        args_dict: dict[str, Any],
        default_job_name: str,
    ) -> "GricadOarsub":
        return cls(
            job_name=args_dict.get("job_name", default_job_name),
            num_nodes=args_dict.get("num_nodes", 1),
            num_cpus=args_dict.get("num_cpus", 0),
            num_cpu_cores=args_dict.get("num_cpu_cores", 0),
            num_gpus=args_dict.get("num_gpus", 1),
            gpu_model=args_dict.get("gpu_model", ""),
            walltime=args_dict.get("walltime", cls._default_walltime),
            log_file_name=args_dict.get("log_file_name", ""),
        )


@dataclass
class Gricad:
    username: str
    project_name: str
    prefered_cluster: str
    project_remote_path: str
    use_container: bool
    singularity_image: str
    oarsub: GricadOarsub

    @classmethod
    def from_dict(
        cls,
        args_dict: dict[str, Any],
        default_job_name: str,
    ) -> "Gricad":
        oarsub_dict: dict[str, Any] = args_dict.get("oarsub", {})
        oarsub: GricadOarsub = GricadOarsub.from_dict(
            args_dict=oarsub_dict,
            default_job_name=default_job_name,
        )

        return cls(
            username=args_dict.get("username"),
            project_name=args_dict.get("project_name"),
            project_remote_path=args_dict.get("project_remote_path"),
            prefered_cluster=args_dict.get("prefered_cluster", "bigfoot"),
            use_container=args_dict.get("use_container", True),
            singularity_image=args_dict.get("singularity_image", "container.sif"),
            oarsub=oarsub,
        )


@dataclass
class RemoteServer(ABC):
    port: int
    open_browser: bool


@dataclass
class TensorBoard(RemoteServer):
    logdir: str

    @classmethod
    def from_dict(cls, args_dict: dict[str, Any]) -> "TensorBoard":
        return cls(
            port=args_dict.get("port", 9090),
            logdir=args_dict.get("logdir", "output/"),
            open_browser=args_dict.get("open_browser", True),
        )


@dataclass
class Jupyter(RemoteServer):
    @classmethod
    def from_dict(cls, args_dict: dict[str, Any]) -> "Jupyter":
        return cls(
            port=args_dict.get("port", 8080), open_browser=args_dict.get("open_browser", True)
        )


@dataclass
class Aim(RemoteServer):
    repo: str

    @classmethod
    def from_dict(
        cls,
        args_dict: dict[str, Any],
    ) -> "Aim":
        return cls(
            port=args_dict.get("port", 43800),
            repo=args_dict.get("repo", "output/"),
            open_browser=args_dict.get("open_browser", True),
        )


@dataclass
class RemoteServers:
    browser_cmd: str
    jupyter: Jupyter
    tensorboard: TensorBoard
    aim: Aim


# pylint: disable=too-few-public-methods
class Config(metaclass=Singleton):
    """
    A class storing all the configuration.
    """

    def __init__(self) -> None:
        config_path: str = join(_DOT_REMI_PATH, _CONFIG_FILE_NAME)
        if not isfile(config_path):
            print_error(
                f"Configuration file (`{config_path}`) was not found ! Exiting."
                "\nYou might want to `cd` to your project or to run"
                " `remi init` if you have not done it yet."
            )
            sys.exit(1)

        with open(config_path, "r") as file_stream:
            config_dict: dict[str, Any] = yaml.load(file_stream.read(), yaml.FullLoader)

        self.initialize_from_dict(config_dict=config_dict)

    def initialize_from_dict(self, config_dict: dict[str, Any]) -> None:
        self.verbose: bool = config_dict.get("verbose", False)

        self.project_name: str = config_dict.get("project_name", basename(abspath(".")))
        self.username: str = config_dict.get("username", "")
        self.exclude_file: str = config_dict.get(
            "exclude_file", join(_DOT_REMI_PATH, "exclude.txt")
        )

        # Desktop
        self.desktop: Desktop = Desktop.from_dict(args_dict=config_dict.get("desktop"))

        self.project_remote_path: str = config_dict.get(
            "project_remote_path",
            "/".join(
                [
                    "/scratch",
                    self.desktop.hostname,
                    self.username,
                    ".remi_projects",
                    self.project_name,
                ]
            ),
        )
        self.working_from_inria: bool
        if gethostname() == self.desktop.hostname:
            self.working_from_inria = True
        else:
            self.working_from_inria = False
        self.output_path: str = "output/"
        self.notebooks_path: str = "notebooks/"
        self.oarsub_log_path: str = join(self.output_path, "oarsub_logs/")

        # Bastion
        self.bastion: Bastion = Bastion.from_dict(
            args_dict=config_dict.get("bastion"),
        )

        # Oarsub
        cluster_oarsub_dict: dict[str, Any] = config_dict.get("oarsub")
        self.oarsub: Oarsub = Oarsub.from_dict(
            args_dict=cluster_oarsub_dict,
            default_job_name=self.project_name,
        )

        # Singularity
        self.singularity: Singularity = Singularity.from_dict(
            args_dict=config_dict.get("singularity")
        )

        # Gricad
        if "gricad" in config_dict:
            self.gricad: Gricad = Gricad.from_dict(
                args_dict=config_dict.get("gricad", {}),
                default_job_name=self.project_name,
            )

        # Remote servers
        remote_servers_dict: dict[str, Any] = config_dict.get("remote_servers", {})

        # Jupyter:
        jupyter: Jupyter = Jupyter.from_dict(args_dict=remote_servers_dict.get("jupyter", {}))

        # TensorBoard:
        tensorboard: TensorBoard = TensorBoard.from_dict(
            args_dict=remote_servers_dict.get("tensorboard", {})
        )

        # Aim:
        aim: Aim = Aim.from_dict(args_dict=remote_servers_dict.get("aim", {}))

        self.remote_servers = RemoteServers(
            browser_cmd=remote_servers_dict.get("browser_cmd", "firefox"),
            jupyter=jupyter,
            tensorboard=tensorboard,
            aim=aim,
        )



"""
CLI utilities for the cluster commands.
"""

import click

from ...config import Config


argument_oar_job_id: click.argument = click.argument(
    "oar-job-id",
    type=int,
    nargs=1,
)

argument_oar_job_id_list: click.argument = click.argument(
    "oar-job-ids",
    type=int,
    nargs=-1,
)
